package org.greenstone.gatherer.file;

import java.io.File;
import java.util.*;
import java.util.regex.*;
import javax.swing.event.*;
import javax.swing.tree.*;
import org.greenstone.gatherer.Gatherer;
import org.greenstone.gatherer.file.FileFilter;
import org.greenstone.gatherer.file.FileNode;
import org.greenstone.gatherer.file.FileOpenActionListener;
import org.greenstone.gatherer.gui.tree.DragTree;
import org.greenstone.gatherer.util.SynchronizedTreeModelTools;

public class FileSystemModel
    extends DefaultTreeModel 
    implements TreeExpansionListener, TreeWillExpandListener {

    private DragTree tree;
    private FileFilter current_filter;
    private FileFilter[] filters;
    /** The filters in place for any file system model. */
    private FileFilter[] default_filters = { new FileFilter("\\..*", true), new FileFilter("metadata\\.xml", true) };

    public FileSystemModel(FileNode root) {
	super(root);
	current_filter = null;
	filters = null;
	tree = null;
	root.setModel(this);
	root.map();
    }

    public FileFilter[] getFilters() {
	if(filters == null) {
	    if(current_filter != null) {
		filters = new FileFilter[default_filters.length + 1];
		filters[default_filters.length] = current_filter;
	    }
	    else {
		filters = new FileFilter[default_filters.length];
	    }
	    System.arraycopy(default_filters, 0, filters, 0, default_filters.length);
	}
	return filters;
    }
	 
    /** Retrieve the node denoted by the given tree path. Note that this isn't equivelent to saying path.lastPathComponent, as the references within the path may be stale. */
    public FileNode getNode(TreePath path) {
	///ystem.err.println("**** getNode(" + path + ") ****");
	FileNode current = (FileNode)root;
	// Special case for the root node. Check the first path component is the root node.
	FileNode first_node = (FileNode)path.getPathComponent(0);
	if(current.equals(first_node)) {
				///ystem.err.println("First path component matches root node.");
				// For each path with this tree path
	    for(int i = 1; current != null && i < path.getPathCount(); i++) {
		// Retrieve the stale path
		FileNode stale_node = (FileNode) path.getPathComponent(i);
		///ystem.err.print("Searching for '" + stale_node + "': ");
		// Locate the fresh node by searching current's children. Remember to ensure that current is mapped.
		current.map();
		boolean found = false;
		for(int j = 0; !found && j < current.getChildCount(); j++) {
		    FileNode child_node = (FileNode) current.getChildAt(j);
		    ///ystem.err.print(child_node + " ");
		    if(stale_node.equals(child_node)) {
			found = true;
			current = child_node;
			///ystem.err.println("Found!");
		    }
		    child_node = null;
		}
		// If no match is found, then set current to null and exit.
		if(!found) {
		    current = null;
		    ///ystem.err.println("Not Found!");
		}
		// Repeat as necessary
	    }
	}
	///ystem.err.println("Returning node: " + new TreePath(current.getPath()));
	return current;
    }

    public void insertNodeInto(MutableTreeNode newChild, MutableTreeNode parent, int index) {
	///ystem.err.println("insertNodeInto(" + newChild + ", " + parent + ", " + index + ")");
	super.insertNodeInto(newChild, parent, index);
    }

    public void mapDirectory(File directory, String title) {
	FileNode node = new FileNode(directory, this, title, true);
	SynchronizedTreeModelTools.insertNodeInto(this, (FileNode)root, node);
    }

    /** Used to refresh the contents of the FileNode indicated by the tree path. While the refresh itself is a piece of the proverbial, the restoring of expanded folders afterwards is a nightmare, mainly because all of the tree paths are no longer valid (remember that refresh involves throwing away the node currents children and remapping them).
     * @param path The TreePath to the node to be refreshed.
     */
    public void refresh(TreePath path) {
	// If no path is set, take the path to the root node (ie update the whole tree)
	if(path == null) {
	    ///ystem.err.println("Refresh entire tree.");
	    path = new TreePath(((FileNode)root).getPath());
	}
	//else {
	//    System.err.println("Refresh: " + path.getLastPathComponent());
	//}
	// Only a valid action if this model is currently being displayed in a tree.
	if(tree != null) {
	    // Retrieve the error node.
	    FileNode node = (FileNode) path.getLastPathComponent();
	    // If this error node is a dummy node (ie has no associated file) we can't unmap it, so we iterate through its children refreshing each in turn. The exception being Greenstone Collections, as it is a dummy node but we can map/unmap it
	    if(node.getFile() == null && !node.toString().equals(Gatherer.dictionary.get("Tree.World"))) {
		for(int i = 0; i < node.getChildCount(); i++) {
		    FileNode child = (FileNode) node.getChildAt(i);
		    refresh(new TreePath(child.getPath()));
		    child = null;
		}
	    }
	    // Otherwise we refresh this node.
	    else {
		// Record all of the expanded paths under this node. How come getExpandedDescendants returns more results each time.
		Enumeration old_tree_paths = tree.getExpandedDescendants(path);
		// Refresh the tree structure.
		node.unmap();
		node.map();
		// Fire the appropriate event.
		nodeStructureChanged(node);
		// Then (painfully) restore the expanded paths. Note that the paths stored in the enumeration no longer exist, so I have to restore by node matching.
		//counter = 0;
		///ystem.err.println("After:");
		while(old_tree_paths != null && old_tree_paths.hasMoreElements()) {
		    //counter++;
		    FileNode current_node = node;
		    TreePath old_tree_path = (TreePath) old_tree_paths.nextElement();
		    ///ystem.err.println(counter + ". " + old_tree_path);
		    TreePath new_tree_path = new TreePath(path.getPath()); // Why isn't there a treepath copy constructor!
		    boolean not_found = false;
		    while(!not_found && old_tree_path.getPathCount() > new_tree_path.getPathCount()) {
			// Retrieve the new node in the old tree path
			FileNode old_path_node = (FileNode) old_tree_path.getPathComponent(new_tree_path.getPathCount());
			// Ensure the current node is mapped
			current_node.map();
			// Now attempt to match it to a child of the new paths last node.
			boolean found = false;
			for(int i = 0; !found && i < current_node.getChildCount(); i++) {
			    FileNode target = (FileNode) current_node.getChildAt(i);
			    ///ystem.err.println("Comparing " + old_path_node.toString() + " with " + target);
			    if(target.toString().equals(old_path_node.toString())) {
				current_node = target;
				found = true;
			    }
			    target = null;
			}
			old_path_node = null;
			// If we found a match, we add that to the new Tree Path and continue.
			if(found) {
			    ///ystem.err.println("Found a match!");
			    new_tree_path = new_tree_path.pathByAddingChild(current_node);
			}
			// We also have to record if we were unable to find a match in the current nodes children.
			else {
			    ///ystem.err.println("Node not found.");
			    not_found = true;
			}
		    }
		    old_tree_path = null;
		    // If we've got this far, and haven't hit a not found, then we can assume we have the appropriate path, and ask the program to expand the new tree path
		    if(!not_found) {
			tree.expandPath(new_tree_path);
			///ystem.err.println(" -> Expanded " + new_tree_path);
		    }
		    else {
			///ystem.err.println(" -> Cannot Expand " + new_tree_path);
		    }
		    new_tree_path = null;
		    current_node = null;
		}
		node = null;
	    }
	}
	else {
	    ///ystem.err.println("No Tree!");
	}
    }

    public void setFilter(String pattern) {
	if(pattern != null) {
	    current_filter = new FileFilter(pattern, false);
	}
	else {
	    current_filter = null;
	}
	filters = null;
    }

    public void setPermanentFilter(String pattern) {
	if(pattern != null) {
	    FileFilter new_filter = new FileFilter(pattern, false);
	    FileFilter temp[] = new FileFilter[default_filters.length + 1];
	    System.arraycopy(default_filters, 0, temp, 0, default_filters.length);
	    temp[default_filters.length] = new_filter;
	    default_filters = temp;
	    temp = null;
	}
	filters = null;
    }

    public void setTree(DragTree tree) {
	this.tree = tree;
    }

    public String toString() {
	if(tree != null) {
	    return tree.toString();
	}
	return "FileSystemModel";
    }

    /** Called whenever an item in the tree has been collapsed. */
    public void treeCollapsed(TreeExpansionEvent event) {
	// Deallocate the affected nodes children. Don't need to do this in a swing worker, as the nodes children are currently not visable.
	TreePath path = event.getPath();
	FileNode node = (FileNode) path.getLastPathComponent();
	node.unmap();
	// Fire the appropriate event.
	nodeStructureChanged(node);
    }

    /** Called whenever an item in the tree has been expanded. */
    public void treeExpanded(TreeExpansionEvent event) {
    }

    /** Invoked whenever a node in the tree is about to be collapsed. */
    public void treeWillCollapse(TreeExpansionEvent event) 
	throws ExpandVetoException {
	// Veto the event if the user is attempting to collapse the root node (regardless of whether it is visible).
	TreePath path = event.getPath();
	if(path.getPathCount() == 1) {
	    throw new ExpandVetoException(event, "Cannot collapse root node!");
	}
    }
	 
    /** Invoked whenever a node in the tree is about to be expanded. */
    public void treeWillExpand(TreeExpansionEvent event) 
	throws ExpandVetoException {
	// Set the wait cursor.
	Gatherer.g_man.wait(true);
	// Allocate the children. Don't need to do this in a swing worker, as the nodes children are currently not visable.
	TreePath path = event.getPath();
	FileNode node = (FileNode) path.getLastPathComponent();
	node.map();
	nodeStructureChanged(node);
	// Restore the cursor.
	Gatherer.g_man.wait(false);
    }
}
