/*
 *    GS2Search.java
 *    Copyright (C) 2002 New Zealand Digital Library, http://www.nzdl.org
 *
 *    This program is free software; you can redistribute it and/or modify
 *    it under the terms of the GNU General Public License as published by
 *    the Free Software Foundation; either version 2 of the License, or
 *    (at your option) any later version.
 *
 *    This program is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU General Public License for more details.
 *
 *    You should have received a copy of the GNU General Public License
 *    along with this program; if not, write to the Free Software
 *    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */
package org.greenstone.gsdl3.service;


// Greenstone classes
import org.greenstone.gdbm.*;
import org.greenstone.gsdl3.util.*;

// XML classes
import org.w3c.dom.Element; 
import org.w3c.dom.Document;
import org.w3c.dom.NodeList; 


/**
 *
 * @author <a href="mailto:mdewsnip@cs.waikato.ac.nz">Michael Dewsnip</a>
 * @version $Revision: 5126 $
 */

public abstract class GS2Search 
    extends ServiceRack {

    // the services on offer
    // these strings must match what is found in the properties file
    protected static final String TEXT_QUERY_SERVICE = "TextQuery";

    // Parameters used
    protected static final String INDEX_PARAM = "index";
    protected static final String CASE_PARAM = "case";
    protected static final String STEM_PARAM = "stem";
    protected static final String MATCH_PARAM = "matchMode";
    protected static final String MATCH_PARAM_ALL = "all";
    protected static final String MATCH_PARAM_SOME = "some";
    protected static final String MAXDOCS_PARAM = "maxDocs";
    protected static final String BOOLEAN_PARAM_ON = "1";
    protected static final String BOOLEAN_PARAM_OFF = "0";
    protected static final String QUERY_PARAM = "query";

    protected static final String EQUIV_TERM_ELEM = "equivTerm";

    protected static final String STEM_ATT = "stem";
    protected static final String NUM_DOCS_MATCH_ATT = "numDocsMatch";
    protected static final String FREQ_ATT = "freq";

    // Elements used in the config file that are specific to this class
    protected static final String DEFAULT_INDEX_ELEM = "defaultIndex";
    protected static final String DEFAULT_LEVEL_ELEM = "defaultLevel";
    protected static final String INDEX_ELEM = "index";
    protected static final String LEVEL_ELEM = "level";

    protected GDBMWrapper gdbm_src = null;

    protected Element config_info = null;

    /** the default index */
    protected String default_index = null;


    /** constructor */
    public GS2Search()
    {
	this.gdbm_src = new GDBMWrapper();
    }


    /** configure this service */
    public boolean configure(Element info, Element extra_info)
    {
	System.out.println("Configuring GS2Search...");
	addExtraQueryInfo(info, extra_info);
	this.config_info = info;

	// Get the default index out of <defaultIndex> (buildConfig.xml)
	Element def = (Element) GSXML.getChildByTagName(info, DEFAULT_INDEX_ELEM);
	if (def != null) {
	    this.default_index = def.getAttribute(GSXML.NAME_ATT);
	}
	if (this.default_index == null || this.default_index.equals("")) {
	    System.err.println("Error: default index not specified!");
	    return false;
	}

	// these entries should reflect the build config file - some services may not be available depending on how the collection was built.
	// set up short_service_info_ - for now just has id and type. the name (lang dependent) will be added in if the list is requested.
	Element tq_service = this.doc.createElement(GSXML.SERVICE_ELEM);
	tq_service.setAttribute(GSXML.TYPE_ATT, GSXML.SERVICE_TYPE_QUERY);
	tq_service.setAttribute(GSXML.NAME_ATT, TEXT_QUERY_SERVICE);
	this.short_service_info.appendChild(tq_service);


	// Open GDBM database for querying
	String gdbm_db_file = GSFile.GDBMDatabaseFile(this.site_home, this.cluster_name);
	if (!this.gdbm_src.openDatabase(gdbm_db_file, GDBMWrapper.READER)) {
	    System.err.println("Error: Could not open GDBM database!");
	    return false;
	}

	// add some format info to service map if there is any
	Element format = (Element) GSXML.getChildByTagName(info, GSXML.FORMAT_ELEM);
	if (format != null) {
	    this.format_info_map.put(TEXT_QUERY_SERVICE, this.doc.importNode(format, true));
	}
	 	
	return true;
    }

    protected Element getServiceDescription(String service, String lang, String subset) {
	
	if (!service.equals(TEXT_QUERY_SERVICE)) {
	    return null;
	}
	Element tq_service = this.doc.createElement(GSXML.SERVICE_ELEM);
	tq_service.setAttribute(GSXML.TYPE_ATT, GSXML.SERVICE_TYPE_QUERY);
	tq_service.setAttribute(GSXML.NAME_ATT, TEXT_QUERY_SERVICE);
	if (subset==null || subset.equals(GSXML.DISPLAY_TEXT_ELEM)) {
	    tq_service.appendChild(GSXML.createDisplayTextElement(this.doc, GSXML.DISPLAY_TEXT_NAME, getTextString(TEXT_QUERY_SERVICE+".name", lang)));
	    tq_service.appendChild(GSXML.createDisplayTextElement(this.doc, GSXML.DISPLAY_TEXT_SUBMIT, getTextString(TEXT_QUERY_SERVICE+".submit", lang)));
	    tq_service.appendChild(GSXML.createDisplayTextElement(this.doc, GSXML.DISPLAY_TEXT_DESCRIPTION, getTextString(TEXT_QUERY_SERVICE+".description", lang)));
	}
	if (subset==null || subset.equals(GSXML.PARAM_ELEM+GSXML.LIST_MODIFIER)) {
	    Element param_list = this.doc.createElement(GSXML.PARAM_ELEM+GSXML.LIST_MODIFIER);
	    createTextQueryParamList(param_list, lang);
	    tq_service.appendChild(param_list);
	}
	return tq_service;
	
    }
    
    protected boolean addExtraQueryInfo(Element info, Element extra_info){

	if (extra_info == null) {
	    return false;
	}
	
	Document owner = info.getOwnerDocument();
	// so far we have index specific display elements, and global format elements 
	NodeList indexes = info.getElementsByTagName(GSXML.INDEX_ELEM);
	Element config_search = (Element)GSXML.getChildByTagName(extra_info, GSXML.SEARCH_ELEM);
	
	for (int i=0; i<indexes.getLength();i++) {
	    Element ind = (Element)indexes.item(i);
	    String name = ind.getAttribute(GSXML.NAME_ATT);
	    Element node_extra = GSXML.getNamedElement(config_search,
						       GSXML.INDEX_ELEM,
						       GSXML.NAME_ATT,
						       name);
	    if (node_extra == null) {
		System.err.println("GS2Search: haven't found extra info for index named "+name);
		continue;
	    }
	    
	    // get the display elements if any - displayName
	    NodeList display_names = node_extra.getElementsByTagName(GSXML.DISPLAY_TEXT_ELEM);
	    if (display_names !=null) {
		for (int j=0; j<display_names.getLength(); j++) {
		    Element e = (Element)display_names.item(j);
		    ind.appendChild(owner.importNode(e, true));
		}
	    }
	} // for each index
	
	// get the format element if any 
	Element format = (Element)GSXML.getChildByTagName(config_search,
							  GSXML.FORMAT_ELEM);
	if (format!=null) { // append to  info
	    info.appendChild(owner.importNode(format, true));
	}
	return true;


    }
/** creates a new param element and adds it to the param list */
    protected void createParameter(String name, Element param_list, 
				   /*boolean display,*/ String lang)
    {
	Element param=null;
	
	if (name.equals(INDEX_PARAM)) {
	    // the index info - read from config file
	    Element index_list = (Element)GSXML.getChildByTagName(this.config_info, INDEX_ELEM+GSXML.LIST_MODIFIER);
	    NodeList indexes = index_list.getElementsByTagName(INDEX_ELEM);
	    int len = indexes.getLength();
	    // now add even if there is only one
	    String [] inds  = new String[len];
	    String [] ind_names = new String[len];
	    for (int i=0; i<len; i++) {
		Element index = (Element)indexes.item(i);
		inds[i] = index.getAttribute(GSXML.NAME_ATT);
		ind_names[i] = GSXML.getDisplayText(index, GSXML.DISPLAY_TEXT_NAME, lang, "en");
		
	    }
	    param = GSXML.createParameterDescription(this.doc, INDEX_PARAM, getTextString("param."+INDEX_PARAM, lang), GSXML.PARAM_TYPE_ENUM_SINGLE, this.default_index, inds, ind_names);
	    
	} 
	else if (name.equals(CASE_PARAM) || name.equals(STEM_PARAM)) {
	    String[] bool_ops = {"0", "1"};
	    String[] bool_texts = {getTextString("param.boolean.off", lang),getTextString("param.boolean.on", lang)}; 
	    param = GSXML.createParameterDescription(this.doc, name, getTextString("param."+name, lang), GSXML.PARAM_TYPE_BOOLEAN, BOOLEAN_PARAM_ON, bool_ops, bool_texts);
	} 
	else if (name.equals(MATCH_PARAM)) {
	    String[] vals = {MATCH_PARAM_ALL, MATCH_PARAM_SOME};
	    String[] val_texts = {getTextString("param."+MATCH_PARAM+"."+MATCH_PARAM_ALL, lang),getTextString("param."+MATCH_PARAM+"."+MATCH_PARAM_SOME, lang)}; 
	    param = GSXML.createParameterDescription(this.doc, MATCH_PARAM, getTextString("param."+MATCH_PARAM, lang), GSXML.PARAM_TYPE_ENUM_SINGLE, MATCH_PARAM_ALL, vals, val_texts);
	    
	}
	else if (name.equals(MAXDOCS_PARAM)) {
	    param = GSXML.createParameterDescription(this.doc, MAXDOCS_PARAM, getTextString("param."+MAXDOCS_PARAM, lang), GSXML.PARAM_TYPE_INTEGER, "10", null, null);
	    
	}
	else if (name.equals(QUERY_PARAM)) {
	    param = GSXML.createParameterDescription(this.doc, QUERY_PARAM, getTextString("param."+QUERY_PARAM, lang), GSXML.PARAM_TYPE_STRING, null, null, null);
	    
	}
	
	// add the param to the list
	if (param != null) {
	    param_list.appendChild(param);
	}
    }


    /** this creates all the params and appends them to param_list.
     * if display=true it creates the text strings version
     * otherwise it creates the description version
     */
    protected abstract boolean createTextQueryParamList(Element param_list, 
							String lang);

    /** Creates a new documentNode element containing ID, node type 
     * and docType*/
    protected Element createDocumentNodeElement(String node_id)
    {
	Element doc_node = this.doc.createElement(GSXML.DOC_NODE_ELEM);
	doc_node.setAttribute(GSXML.NODE_ID_ATT, node_id);

	String top_id = OID.getTop(node_id);
	boolean is_top = (top_id.equals(node_id) ? true : false);
	
	DBInfo info = this.gdbm_src.getInfo(node_id);
	if (info == null) { // make it up - cant query the gdbm db
	    doc_node.setAttribute(GSXML.DOC_TYPE_ATT, "simple");
	    return doc_node;
	}
	String children = info.getInfo("contains");
	boolean is_leaf = (children.equals("") ? true : false);

	// check for simple doc types
	if (is_top && is_leaf) { // a single section document
	    doc_node.setAttribute(GSXML.DOC_TYPE_ATT, "simple");
	    return doc_node;
	}
	// set teh node type att
	if (is_top) {
	    doc_node.setAttribute(GSXML.NODE_TYPE_ATT, GSXML.NODE_TYPE_ROOT);
	} else if (is_leaf) {
	    doc_node.setAttribute(GSXML.NODE_TYPE_ATT, GSXML.NODE_TYPE_LEAF);
	} else {
	    doc_node.setAttribute(GSXML.NODE_TYPE_ATT, GSXML.NODE_TYPE_INTERIOR);
	}
	
	if (!is_top) { // we need to look at the top info
	    info = this.gdbm_src.getInfo(top_id);
	}

	String childtype = info.getInfo("childtype");
	if (childtype.equals("Paged")) {
	    doc_node.setAttribute(GSXML.DOC_TYPE_ATT, "paged");
	} else {
	    doc_node.setAttribute(GSXML.DOC_TYPE_ATT, "hierarchy");
	}
	return doc_node;
    }
    
    /** returns the document type of a node - if the node is a subnode, it returns teh type of teh top document node. for now, only does paged and hierarchical, but eventually will have more types */
    protected String getDocType(String node_id) {

	String doc_id = OID.getTop(node_id);
	DBInfo info = this.gdbm_src.getInfo(doc_id);
	String child_type = info.getInfo("childtype");
	if (child_type.equals("Paged")) {
	    return "paged";
	}
	return "hierarchy";
	
    }

    /** Returns true if the OID specifies a leaf node, false otherwise
	Note: this makes a request to the GDBM database so it may not be cheap */
    protected boolean isLeafNode(String oid)
    {
	DBInfo info = this.gdbm_src.getInfo(oid);
	String children = info.getInfo("contains");
	return (children.equals(""));
    }


    /** Process a text query - implemented by concrete subclasses */
    protected abstract Element processTextQuery(Element request);
}
